from .seqclasses import _genvals, EndPairArrayTD, EndPairArrayDT, BoolSeqArray
from .energetics_daoe import EnergeticsDAOEC
from .filters import SeqFilter
from .choosers import BasicChooser

import numpy as np

def genpre(endtype, endlength, energetics=None, seqfilter=None, interaction=None,
           oldends=None):
    if energetics is None:
        energetics = EnergeticsDAOEC(endlength)

    epc = {'TD': EndPairArrayTD, 'DT': EndPairArrayDT}[endtype]

    availends = epc(_genvals(BoolSeqArray('n' * (endlength + 2))))
        
    if interaction is None:
        if oldends is not None:
            interaction = np.mean(energetics.gse(oldends))
        else:
            interaction = np.median(energetics.gse(availends))

    if seqfilter is None:
        seqfilter = SeqFilter(energetics, interaction)

    availends = seqfilter.filterspace(availends)

    if oldends is not None:
        availends = seqfilter.filterseqs(availends, oldends)
        
    return availends, interaction, seqfilter
    

def easyends(endtype, endlength, energetics=None, number=None,
             seqfilter=None, chooser=None, preavail=None, interaction=None,
             sfopts={}):

    if energetics is None:
        energetics = EnergeticsDAOEC(endlength)

    if preavail is None:
        epc = {'TD': EndPairArrayTD, 'DT': EndPairArrayDT}[endtype]

        availends = epc(_genvals(BoolSeqArray('n' * (endlength + 2))))
        
        if interaction is None:
            interaction = np.median(energetics.gse(availends))

        if seqfilter is None:
            seqfilter = SeqFilter(energetics, interaction, **sfopts)
        
        availends = seqfilter.filterspace(availends)
    else:
        availends = preavail.copy()
        
    if chooser is None:
        chooser = BasicChooser(energetics, interaction, wiggle=0.4)

    

    curends = None
    while True:
        newend = chooser.choose(availends, curends)
        if curends is None:
            curends = newend[None, :]
        else:
            curends = np.concatenate((curends, newend[None, :]),
                                     0).view(newend.__class__)
        availends = seqfilter.filterseqs(availends, newend[None, :])
        if len(availends) == 0:
            break
        elif number and len(curends) >= number:
            break

    return curends



import numpy as np


from .stickydesign import energy_array_uniform


def hist_multi(all_ends,
               all_energetics,
               energetics_names=None,
               title="", **kwargs):
    from matplotlib import pylab
    fig = pylab.figure(figsize=(10, 15))
    a, b, c = fig.subplots(3, 1)
    a.hist(
        [
            np.concatenate(tuple(en.matching_uniform(y) for y in all_ends))
            for en in all_energetics
        ],
        bins=50,
        label=energetics_names,
        **kwargs)
    if energetics_names:
        a.legend()
    a.set_xlabel("$ΔG_{se}$ (kcal/mol)")
    a.set_ylabel("# of interactions")
    a.set_title("Matching ends")

    b.hist(
        [
            np.concatenate(
                tuple(
                    np.ravel(energy_array_uniform(y, en)) for y in all_ends))
            for en in all_energetics
        ],
        bins=100,
        label=energetics_names, **kwargs)
    b.set_xlabel("$ΔG_{se}$ (kcal/mol)")
    b.set_ylabel("# of interactions")
    b.set_title("All ends")
    bins = np.linspace(2.5, 9.0, 100)

    c.hist(
        [
            np.concatenate(
                tuple(
                    np.ravel(energy_array_uniform(y, en)) for y in all_ends))
            for en in all_energetics
        ],
        bins=bins,
        label=energetics_names, **kwargs)
    c.set_xlim(2.5, 9.0)
    c.set_xlabel("$ΔG_{se}$ (kcal/mol)")
    c.set_ylabel("# of interactions")
    c.set_title("All ends (zoomed)")
    if title:
        fig.suptitle(title)
    fig.tight_layout(rect=[0, 0.3, 1, 0.97])
    return fig

def box_multi(all_ends,
              all_energetics,
              energetics_names=None,
              title="", **kwargs):
    from matplotlib import pylab
    fig = pylab.figure(figsize=(10, 15))
    a = fig.subplots(1, 1)
    a.boxplot(
        [
            np.concatenate(tuple(en.matching_uniform(y) for y in all_ends))
            for en in all_energetics
        ],
        labels=energetics_names, sym='r.',
        **kwargs)
    a.set_ylabel("$ΔG_{se}$ (kcal/mol)")

    v = []
    for en in all_energetics:
        l = []
        for y in all_ends:
            x = np.ma.masked_array(energy_array_uniform(y, en))
            i = np.arange(0,len(x)//2)
            x[i,len(x)//2+i]=np.ma.masked
            x[np.tril_indices_from(x,-1)]=np.ma.masked
            l.append(x.compressed())
        v.append(np.concatenate(tuple(l)))
    
    a.boxplot(v, sym='bP', labels=energetics_names, **kwargs)

    if title:
        fig.suptitle(title)
    fig.tight_layout(rect=[0, 0.3, 1, 0.97])
    return fig


def heatmap(ends, energetics, title="", **kwargs):
    from matplotlib import pylab
    fig = pylab.figure()

    heat = energy_array_uniform(ends, energetics)

    pylab.imshow(heat)
    pylab.colorbar()
    pylab.title(title)

    return fig


def _multi_data_pandas(ends, all_energetics, energetics_names=None):
    """From a list of endarrays, make two Pandas DataFrames: one of
    matching energies, the other of spurious/mismatch energies.

    Returns
    -------

    match : pandas.DataFrame
        matching energies

    mismatch: pandas.DataFrame
        mismatch energies
    """
    import pandas as pd
    match = np.array([np.concatenate(
        [e.matching_uniform(x) for x in ends]) for e in all_energetics]).T
    match = pd.DataFrame(match, columns=energetics_names)
    
    v = []
    for en in all_energetics:
        l = []
        for y in ends:
            x = np.ma.masked_array(energy_array_uniform(y, en))
            i = np.arange(0, len(x)//2)
            x[i, len(x)//2+i] = np.ma.masked
            x[np.tril_indices_from(x, -1)] = np.ma.masked
            l.append(x.compressed())
        v.append(np.concatenate(tuple(l)))
    return match, pd.DataFrame(np.array(v).T, columns=energetics_names)



from __future__ import division
import numpy as np
from .endclasses import pairseqa, tops, endarray, Energetics
from .version import __version__

from . import newparams as p


class EnergeticsDAOE(Energetics):
    """Energy functions based on several sources, primarily SantaLucia's
    2004 paper, along with handling of dangles, tails, and nicks
    specifically for DX tile sticky ends.

    
    Parameters
    ----------
    coaxparams : str or False, optional
        choose the coaxial stacking parameters to use.  False is no coaxial
        stacking adjustment, other options are 'protozanova', 'peyret',
        and 'pyshni'.

    singlepair : bool, optional
        Treat single base pair pairing as possible (defaults
        to False).

    temperature : float, optional
        Temperature to use for the model, in degress Celsius (defaults to 37).

    """

    def __init__(self,
                 temperature=37,
                 mismatchtype=None,
                 coaxparams=False,
                 singlepair=False,
                 danglecorr=True,
                 version=None,
                 enclass=None):
        self.coaxparams = coaxparams
        self.singlepair = singlepair
        self.danglecorr = danglecorr
        self.temperature = temperature

    @property
    def info(self):
        info = {'enclass': 'EnergeticsDAOE',
                'temperature': self.temperature,
                'coaxparams': self.coaxparaminfo,
                'danglecorr': self.danglecorr,
                'singlepair': self.singlepair,
                'version': __version__}
        return info

    @property
    def temperature(self):
        return self._temperature

    @temperature.setter
    def temperature(self, val):
        self._temperature = val
        self._setup_params(val)
    
    def __str__(self):
        return "EnergeticsDAOE(" + \
            ", ".join("{}={}".format(x, repr(y))
                      for x, y in self.info.items()) + \
            ")"

    def __repr__(self):
        return self.__str__()
    
    def _setup_params(self, temperature=37):
        self.initdG = p.initdG37 - (temperature - 37) * p.initdS
        self.nndG = p.nndG37 - (temperature - 37) * p.nndS
        if self.coaxparams == 'protozanova' or self.coaxparams is True:
            self.coaxddG = p.coaxddG37 - (temperature - 37) * p.coaxddS
            self.coaxparaminfo = 'protozanova'
            self.coaxparams = True
        elif self.coaxparams == 'peyret':
            self.coaxddG = p.coax_peyret_ddG37 - (
                temperature - 37) * p.coax_peyret_ddS
            self.coaxparaminfo = self.coaxparams
            self.coaxparams = True
        elif self.coaxparams == 'pyshni':
            self.coaxddG = p.coax_pyshni_ddG37 - (
                temperature - 37) * p.coax_pyshni_ddS
            self.coaxparaminfo = self.coaxparams
            self.coaxparams = True
        elif not self.coaxparams:
            self.coaxddG = np.zeros_like(p.coaxddG37)
            self.coaxparaminfo = self.coaxparams
            self.coaxparams = False
        else:
            raise ValueError("Invalid coaxparams: {}".format(self.coaxparams),
                             self.coaxparams)
        self.dangle5dG = p.dangle5dG37 - (temperature - 37) * p.dangle5dS
        self.dangle3dG = p.dangle3dG37 - (temperature - 37) * p.dangle3dS
        self.intmmdG = p.intmmdG37 - (temperature - 37) * p.intmmdS

        self.ltmmdG_5335 = np.zeros(256)
        self.rtmmdG_5335 = np.zeros(256)
        self.intmmdG_5335 = np.zeros(256)

        # Dumb setup. FIXME: do this cleverly
        for i in range(0, 4):
            for j in range(0, 4):
                for k in range(0, 4):
                    self.ltmmdG_5335[
                        i * 64 + j * 16 + k * 4 +
                        j] = self.dangle5dG[i * 4
                                            + j] + self.dangle3dG[(3 - j) * 4
                                                                  + (3 - k)]
                    self.rtmmdG_5335[
                        i * 64 + j * 16 + i * 4 +
                        k] = self.dangle3dG[i * 4
                                            + j] + self.dangle5dG[(3 - k) * 4
                                                                  + (3 - i)]
                    self.intmmdG_5335[i * 64 + j * 16 + k * 4 +
                                      j] = self.intmmdG[(3 - j) * 16
                                                        + (3 - k) * 4 + i]
                    self.intmmdG_5335[i * 64 + j * 16 + i * 4 +
                                      k] = self.intmmdG[i * 16
                                                        + j * 4 + (3 - k)]

    def matching_uniform(self, seqs):
        ps = pairseqa(seqs)

        # In both cases here, the energy we want is the NN binding energy of
        # each stack,
        if seqs.endtype == 'DT':
            dcorr = np.zeros_like(self.dangle3dG[ps[:, 0]])
            if self.danglecorr:
                dcorr += -self.dangle3dG[ps[:, 0]] - self.dangle3dG[ps.revcomp(
                )[:, 0]]
            if self.coaxparams:
                dcorr += self.coaxddG[ps[:, 0]] + self.coaxddG[ps.revcomp()[:,
                                                                            0]]
        elif seqs.endtype == 'TD':
            dcorr = np.zeros_like(self.dangle5dG[ps[:, -1]])
            if self.danglecorr:
                dcorr += (-self.dangle5dG[ps[:, -1]] -
                          self.dangle5dG[ps.revcomp()[:, -1]])
            if self.coaxparams:
                dcorr += self.coaxddG[ps[:, -1]] + self.coaxddG[ps.revcomp()
                                                                [:, -1]]
        else:
            raise NotImplementedError
        return -(np.sum(self.nndG[ps], axis=1) + self.initdG + dcorr)

    def uniform(self, seqs1, seqs2, debug=False):
        if seqs1.shape != seqs2.shape:
            if seqs1.ndim == 1:
                seqs1 = endarray(
                    np.repeat(np.array([seqs1]), seqs2.shape[0], 0),
                    seqs1.endtype)
            else:
                raise ValueError(
                    "Lengths of sequence arrays are not acceptable.")

        assert seqs1.endtype == seqs2.endtype
        endtype = seqs1.endtype

        s1 = tops(seqs1)
        s2 = tops(seqs2)
        l = s1.shape[1]

        # s2r is revcomp pairseq of s2.
        s2r = np.fliplr(np.invert(s2) % 16)
        s2r = s2r // 4 + 4 * (s2r % 4)

        alloffset_max = np.zeros(
            s1.shape[0])  # store for max binding at any offset

        if endtype == 'TD':
            s1_end = s1[:, 0:-1]  #
            s2_end_rc = s2r[:, 1:]
            s1l = np.hstack(((4 * (s2r[:, 0] // 4) + s1[:, 0] // 4).reshape(
                -1, 1), s1))
            s2rl = np.hstack(
                (s2r, (4 * (s2r[:, -1] % 4) + s1[:, -1] % 4).reshape(-1, 1)))
        elif endtype == 'DT':
            s1_end = s1[:, 1:]
            s2_end_rc = s2r[:, 0:-1]
            s2rl = np.hstack(((4 * (s1[:, 0] // 4) + s2r[:, 0] // 4).reshape(
                -1, 1), s2r))
            s1l = np.hstack(
                (s1, (4 * (s1[:, -1] % 4) + s2r[:, -1] % 4).reshape(-1, 1)))
        else:
            raise NotImplementedError

        for offset in range(-l + 2, l - 1):
            if offset > 0:
                if endtype == 'TD':
                    # Energies of matching stacks, zero otherwise. Can be used
                    # to check match.
                    ens = (s1_end[:, :-offset] == s2_end_rc[:, offset:]) * (
                        -self.nndG[s1_end[:, :-offset]])
                    ens[:, 0] += (
                        ens[:, 0] != 0) * (-self.dangle3dG[s1_end[:, -offset]]
                                           )  # - for positive sign
                    ens[:, -1] += (
                        ens[:, -1] != 0) * (-self.dangle3dG[s2[:, -offset - 1]]
                                            )  # - for positive sign
                    ltmm = -self.ltmmdG_5335[s1_end[:, :-offset] * 16
                                             + s2_end_rc[:, offset:]]
                    rtmm = -self.rtmmdG_5335[s1_end[:, :-offset] * 16
                                             + s2_end_rc[:, offset:]]
                    intmm = -self.intmmdG_5335[s1_end[:, :-offset] * 16
                                               + s2_end_rc[:, offset:]]
                if endtype == 'DT':
                    ens = (s1_end[:, offset:] == s2_end_rc[:, :-offset]) * (
                        -self.nndG[s1_end[:, offset:]])
                    ens[:, 0] += (ens[:, 0] != 0) * (
                        -self.dangle5dG[s1_end[:, offset - 1]]
                    )  # - for positive sign
                    ens[:, -1] += (ens[:, -1] != 0) * (
                        -self.dangle5dG[s2[:, offset]])  # - for positive sign
                    ltmm = -self.ltmmdG_5335[s1_end[:, offset:] * 16
                                             + s2_end_rc[:, :-offset]]
                    rtmm = -self.rtmmdG_5335[s1_end[:, offset:] * 16
                                             + s2_end_rc[:, :-offset]]
                    intmm = -self.intmmdG_5335[s1_end[:, offset:] * 16
                                               + s2_end_rc[:, :-offset]]
            elif offset == 0:
                ens = (s1_end == s2_end_rc) * (-self.nndG[s1_end])
                if endtype == 'DT':
                    ens[:, 0] += (ens[:, 0] != 0) * (
                        +self.dangle3dG[s1[:, 0]] -
                        self.coaxparams * self.coaxddG[s1[:, 0]]) - (
                            s1l[:, 0] == s2rl[:, 0]
                        ) * self.nndG[s1l[:, 0]]  # sign reversed
                    ens[:, -1] += (ens[:, -1] != 0) * (
                        +self.dangle3dG[s2[:, 0]] -
                        self.coaxparams * self.coaxddG[s2[:, 0]]) - (
                            s1l[:, -1] == s2rl[:, -1]
                        ) * self.nndG[s1l[:, -1]]  # sign reversed
                if endtype == 'TD':
                    ens[:, 0] += +(ens[:, 0] != 0) * (
                        self.dangle5dG[s1[:, -1]] -
                        self.coaxparams * self.coaxddG[s1[:, -1]]) - (
                            s1l[:, 0] == s2rl[:, 0]
                        ) * self.nndG[s1l[:, 0]]  # sign reversed
                    ens[:, -1] += +(ens[:, -1] != 0) * (
                        self.dangle5dG[s2[:, -1]] -
                        self.coaxparams * self.coaxddG[s2[:, -1]]) - (
                            s1l[:, -1] == s2rl[:, -1]
                        ) * self.nndG[s1l[:, -1]]  # sign reversed
                ltmm = np.zeros_like(ens)
                rtmm = np.zeros_like(ens)
                intmm = np.zeros_like(ens)
                ltmm = -self.ltmmdG_5335[s1_end[:, :] * 16 + s2_end_rc[:, :]]
                rtmm = -self.rtmmdG_5335[s1_end[:, :] * 16 + s2_end_rc[:, :]]
                intmm = -self.intmmdG_5335[s1_end[:, :] * 16 + s2_end_rc[:, :]]
            else:  # offset < 0
                if endtype == 'TD':
                    ens = (s1_end[:, -offset:] == s2_end_rc[:, :offset]) * (
                        -self.nndG[s1_end[:, -offset:]])
                    ens[:, 0] += (ens[:, 0] != 0) * (
                        -self.nndG[s1[:, -1]] - p.tailcordG37 +
                        self.dangle5dG[s1[:, -1]])  # - for positive sign
                    ens[:, -1] += (ens[:, -1] != 0) * (
                        -self.nndG[s2[:, -1]] - p.tailcordG37 +
                        self.dangle5dG[s2[:, -1]])  # - for positive sign
                    ltmm = -self.ltmmdG_5335[s1_end[:, -offset:] * 16
                                             + s2_end_rc[:, :offset]]
                    rtmm = -self.rtmmdG_5335[s1_end[:, -offset:] * 16
                                             + s2_end_rc[:, :offset]]
                    intmm = -self.intmmdG_5335[s1_end[:, -offset:] * 16
                                               + s2_end_rc[:, :offset]]
                elif endtype == 'DT':
                    ens = (s1_end[:, :offset] == s2_end_rc[:, -offset:]) * (
                        -self.nndG[s1_end[:, :offset]])
                    ens[:, 0] += (ens[:, 0] != 0) * (
                        -self.nndG[s1[:, 0]] - p.tailcordG37 +
                        self.dangle3dG[s1[:, 0]])  # - for positive sign
                    ens[:, -1] += (ens[:, -1] != 0) * (
                        -self.nndG[s2[:, 0]] - p.tailcordG37 +
                        self.dangle3dG[s2[:, 0]])  # - for positive sign
                    ltmm = -self.ltmmdG_5335[s1_end[:, :offset] * 16
                                             + s2_end_rc[:, -offset:]]
                    rtmm = -self.rtmmdG_5335[s1_end[:, :offset] * 16
                                             + s2_end_rc[:, -offset:]]
                    intmm = -self.intmmdG_5335[s1_end[:, :offset] * 16
                                               + s2_end_rc[:, -offset:]]
            bindmax = np.zeros(ens.shape[0])
            if debug:
                print(offset, ens.view(np.ndarray), ltmm, rtmm, intmm)
            for e in range(0, ens.shape[0]):
                acc = 0
                for i in range(0, ens.shape[1]):
                    if ens[e, i] != 0:
                        # we're matching. add the pair to the accumulator
                        acc += ens[e, i]
                    elif rtmm[e, i] != 0 and i > 0 and ens[e, i-1] > 0:
                        # we're mismatching on the right: see if right-dangling
                        # is highest binding so far, and continue, adding intmm
                        # to accumulator.
                        if acc + rtmm[e, i] > bindmax[e]:
                            bindmax[e] = acc + rtmm[e, i]
                        acc += intmm[e, i]
                    elif ltmm[e, i] != 0 and i < ens.shape[1] - 1:
                        # don't do this for the last pair we're mismatching on
                        # the left: see if our ltmm is stronger than our
                        # accumulated binding+intmm. If so, reset to ltmm and
                        # continue as left-dangling, or reset to 0 if ltmm+next
                        # is weaker than next dangle,or next is also a mismatch
                        # (fixme: good idea?). If not, continue as internal
                        # mismatch.
                        if (not self.singlepair) and (ltmm[e, i] >
                                                      acc + intmm[e, i]) and (
                                                          ens[e, i + 1] > 0):
                            acc = ltmm[e, i]
                        elif (self.singlepair) and (ltmm[e, i] >
                                                    acc + intmm[e, i]):
                            acc = ltmm[e, i]
                        else:
                            acc += intmm[e, i]
                    else:  # we're at a loop. Add stuff.
                        acc -= p.looppenalty
                bindmax[e] = max(bindmax[e], acc)
            alloffset_max = np.maximum(alloffset_max, bindmax)
        return alloffset_max - self.initdG



from .version import __version__

from .stickydesign import *
from .endclasses import *
from .energetics_daoe import EnergeticsDAOE
from .energetics_basic import EnergeticsBasic
from .energetics_basic_old import EnergeticsBasicOld

# Try to preserve Py2 support for now
# from . import multimodel





import numpy as np

_REVBITS4 = np.array(
    [0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15], dtype=np.uint8)

_BSA_TO_SA: np.ma.MaskedArray = np.ma.masked_array(
    data=[9, 0, 1, 9, 2, 9, 9, 9, 3] + [9] * 7,
    mask=[True, False, False, True, False, True, True, True, False] +
    [True] * 7,
    dtype='uint8')

_BSA_TO_SA_ARRAYS = np.array([
    np.array([0b1 & x, 0b10 & x, 0b100 & x,
              0b1000 & x]).nonzero()[0].astype(np.uint8) for x in range(0, 16)
], dtype=object)

def seqconcat(s1, s2):
    if isinstance(s1, SeqArrayBase) and isinstance(s2, SeqArrayBase):
        return np.concatenate((s1, s2), -1)

class SeqArrayBase(np.ndarray):
    """A base class for numpy-based sequence arrays.  This doesn't assume
    anything about how the array works, and can be used with isinstance
    to check.  Other classes should pull from this."""

    def __new__(cls, array):
        if isinstance(array, np.ndarray):
            return np.asarray(array, dtype=np.uint8).view(cls)            
        if type(array) is str:
            array = [array]
        if type(array[0]) is str:
            array = np.array(
                [[cls._nt[x] for x in y] for y in array], dtype=np.uint8)
        return np.asarray(array, dtype=np.uint8).view(cls)            

    @property
    def strs(self):
        if self.ndim == 2:
            return [''.join(self._l[x] for x in s) for s in self]
        elif self.ndim == 1:
            return ''.join(self._l[x] for x in self)    
    
    def __repr__(self):
        return "{}({})".format(type(self).__name__, repr(self.strs))

_NNRC = np.array([15, 11,  7,  3,
                  14, 10,  6,  2,
                  13,  9,  5,  1,
                  12,  8,  4,  0], dtype=np.uint8)
class NNSeqArray(np.ndarray):
    @property
    def _rccalc(self):
        return 4 * (3 - (self[..., ::-1] % 4)) + 3 - (self[..., ::-1] // 4)

    @property
    def rc(self):
        return _NNRC[self[..., ::-1]].view(self.__class__)

class SeqArray(SeqArrayBase):
    """A SeqArray that uses typical 0,1,2,3 = a,c,g,t coding"""
    _nt = {'a': 0, 'c': 1, 'g': 2, 't': 3}
    _l = {v: k for k, v in _nt.items()}

    @property
    def rc(self):
        return (3 - self)[..., ::-1].view(self.__class__)

    @property
    def nnseq(self):
        return (4 * self[..., :-1] + self[..., 1:]).view(NNSeqArray)

    @property
    def seqlen(self):
        return self.shape[-1]


class BoolSeqArray(SeqArrayBase):
    """A SeqArray that uses the 'bool' coding (a=0b0001, c=0b0010,
    n=0b1111, null=0b0000 etc)"""
    _nt = {
        'a': 1,
        'b': 14,
        'c': 2,
        'd': 13,
        'g': 4,
        'h': 11,
        'k': 12,
        'm': 3,
        'n': 15,
        's': 6,
        't': 8,
        'v': 7,
        'w': 9
    }
    _l = {v: k for k, v in _nt.items()}

    @property
    def rc(self):
        return _REVBITS4[self][..., ::-1].view(self.__class__)

    @property
    def seqarray(self):
        l = _BSA_TO_SA[self]
        if np.ma.is_masked(l):
            raise ValueError("There are ambiguous bases")
        return l.view(SeqArray)


class EndArray(SeqArray):
    pass


class SeqPairArray(SeqArray):
    @property
    def seqs(self):
        return self.view(SeqArray)

    @property
    def pairs(self):
        return self.rc.view(SeqArray)

    comps = pairs

class EndPairArray(EndArray):
    pass


class EndArrayDT(EndArray):
    pass


class EndArrayTD(EndArray):
    pass


class EndPairArrayDT(EndArray):
    @property
    def ends(self):
        """
Returns
-------
EndArrayDT
    an array of just the ends (+ their adjacents) from the array. 

        """
        return self[..., :-1].view(EndArrayDT)

    seqs = ends
    
    @property
    def comps(self):
        """
Returns
-------
EndArrayDT
    an array of just the complementary ends (+ their adjacents) 
    of the ends in the array.
        """
        return self.rc[..., :-1].view(EndArrayDT)

    pairs = comps


class EndPairArrayTD(EndArray):
    @property
    def ends(self):
        """
Returns
-------
EndArrayDT
    an array of just the ends (+ their adjacents) from the array. 

        """
        return self[..., 1:].view(EndArrayTD)

    seqs = ends
    
    @property
    def comps(self):
        """
Returns
-------
EndArrayDT
    an array of just the complementary ends (+ their adjacents) 
    of the ends in the array.
        """
        return self.rc[..., 1:].view(EndArrayTD)

    pairs = comps


def _cartesian_product(*arrays):
    la = len(arrays)
    dtype = np.result_type(*arrays)
    arr = np.empty([len(a) for a in arrays] + [la], dtype=dtype)
    for i, a in enumerate(np.ix_(*arrays)):
        arr[..., i] = a
    return arr.reshape(-1, la)


def _genvals(template: BoolSeqArray):
    return _cartesian_product(*_BSA_TO_SA_ARRAYS[template[0]]).view(SeqArray)



import numpy as np


class Energetics(object):
    pass


class pairseqa(np.ndarray):
    def __new__(cls, array):
        obj = (4 * array[:, :-1] + array[:, 1:]).view(cls)
        return obj

    def revcomp(self):
        return 4 * (3 - (self[:, ::-1] % 4)) + 3 - (self[:, ::-1] // 4)

    def tolist(self):
        st = ["a", "c", "g", "t"]
        return [
            "".join([st[x // 4] for x in y] + [st[y[-1] % 4]]) for y in self
        ]

    def __repr__(self):
        return "{}".format(repr(self.tolist()))


class endarray(np.ndarray):
    """
    This is a class for arrays full ends (of type
    adjacent+end+wc-adjacent-of-complementary-end).

    At present, it also handles adjacent+end style ends, but self.end and
    self.comp will return bogus information. It eventually needs to be split up
    into two classes in order to deal with this problem.
    """

    def __new__(cls, array, endtype):
        if type(array[0]) is str:
            array = np.array(
                [[nt[x] for x in y] for y in array], dtype=np.uint8)
        obj = np.asarray(array, dtype=np.uint8).view(cls)
        obj.endtype = endtype
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.endtype = getattr(obj, 'endtype', None)

    def _get_ends(self):
        if self.endtype == 'DT':
            return self[:, :-1]
        elif self.endtype == 'TD':
            return self[:, 1:]
        elif self.endtype == 'S':
            return self[:, :]

    def _get_comps(self):
        if self.endtype == 'DT':
            return (3 - self)[:, ::-1][:, :-1]
        elif self.endtype == 'TD':
            return (3 - self)[:, ::-1][:, 1:]
        elif self.endtype == 'S':
            return (3 - self)[:, ::-1][:, :]

    @property
    def fcomps(self):
        return (3-self)[:, ::-1]
        
    def concat(a1, a2):
        assert a1.endtype == a2.endtype

        return endarray(
            np.concatenate((np.array(a1), np.array(a2)), axis=0),
            a1.endtype)
        
    def _get_adjs(self):
        if self.endtype == 'DT':
            return self[:, 0]
        elif self.endtype == 'TD':
            return self[:, -1]

    def _get_cadjs(self):
        if self.endtype == 'DT':
            return (3 - self)[:, -1]
        elif self.endtype == 'TD':
            return (3 - self)[:, 0]

    def __len__(self):
        return self.shape[0]

    def _get_endlen(self):
        return self.shape[1] - 1

    def _get_seqlen(self):
        return self.shape[1]

    def append(s1, s2):
        assert s1.endtype == s2.endtype
        n = np.vstack((s1, s2)).view(endarray)
        n.endtype = s1.endtype
        return n

    def __repr__(self):
        return "<endarray ({2}): type {0}; {1}>".format(
            self.endtype, repr(self.tolist()), len(self))

    def tolist(self):
        st = ["a", "c", "g", "t"]
        return ["".join([st[x] for x in y]) for y in self]

    endlen = property(_get_endlen)
    seqlen = property(_get_seqlen)  # Added for new code compatibility
    ends = property(_get_ends)
    comps = property(_get_comps)
    adjs = property(_get_adjs)
    cadjs = property(_get_cadjs)
    strings = property(tolist)


nt = {'a': 0, 'c': 1, 'g': 2, 't': 3}

lton = {
    'a': [0],
    'b': [1, 2, 3],
    'c': [1],
    'd': [0, 2, 3],
    'g': [2],
    'h': [0, 1, 3],
    'k': [2, 3],
    'm': [0, 1],
    'n': [0, 1, 2, 3],
    's': [1, 2],
    't': [3],
    'v': [0, 1, 2],
    'w': [0, 3]
}

wc = {
    'a': 't',
    'b': 'v',
    'c': 'g',
    'd': 'h',
    'g': 'c',
    'h': 'd',
    'k': 'm',
    'm': 'k',
    'n': 'n',
    's': 's',
    't': 'a',
    'v': 'b',
    'w': 'w'
}

def tops(s):
    return 4 * s[:, :-1] + s[:, 1:]



from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext


class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        #__builtins__.__NUMPY_SETUP__ = False
        import numpy # NOQA
        self.include_dirs.append(numpy.get_include())


# define the extension module
stickyext = Extension(
    'stickydesign._stickyext',
    sources=['src/stickydesign/_stickyext.c'])
#    include_dirs=[numpy.get_include()])

# run the setup
setup(
    name='stickydesign',
    version='0.8.4',
    setup_requires=['numpy'],
    install_requires=['numpy'],
    extras_require={'testing': ['pytest-cov', 'pytest']},
    package_dir = {
        'stickydesign': 'src/stickydesign',
        'stickydesign.stickydesign2': 'src/stickydesign/stickydesign2'
        },
    packages=['stickydesign', 'stickydesign.stickydesign2'],
    ext_modules=[stickyext],
    cmdclass={'build_ext':build_ext},
    package_data={'stickydesign': ['params/dnastackingbig.csv']},
    author="Constantine Glen Evans",
    author_email="constantine@costi.net",
    description="StickyDesign DNA Tile Sticky End Package",
    url='http://dna.caltech.edu/StickyDesign',
    zip_safe=False,
    )
